﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Web;
using System.Web.UI;
using System.Web.UI.HtmlControls;
using System.Web.UI.WebControls;
using Netbiz.Bizworks.Data;
using Netbiz.Bizworks.Web;

namespace Netbiz.Bizworks.Web
{
    /// <summary>
    /// Base webPage
    /// </summary>
    /// <typeparam name="TPagePresenter">bw page presenter</typeparam>
    public abstract class BaseWebPage<TPagePresenter> : Page where TPagePresenter : BasePresenter
    {
        #region CONSTRUCTORS
        /// <summary>
        /// constructor blocked.. force derived page to implement
        /// </summary>
        private BaseWebPage() { } // dont allow this ;(
        /// <summary>
        /// Netbiz.Bizworks.Web.BaseWebPage
        /// </summary>
        /// <param name="pPage">key into the bw.model where the page definition is stored</param>
        protected BaseWebPage(string pPage = "unknown")
        {
            _baseContainer = HttpContext.Current.ApplicationInstance.GetBizworksContainer();
            _aspSystemManager = (AspSystemManager)_baseContainer.SystemManager;

            PagePresenter = (TPagePresenter)_baseContainer.Resolve(typeof(TPagePresenter));
            PagePresenter.ResetEvents();
            PagePresenter.RaiseNavigationEvent += PagePresenterRaiseNavigationEvent;
            PageProfile = _aspSystemManager.GetPageProfile(pPage);
            if (PageProfile.PageTitle != null)
            {
                Title = PageProfile.PageTitle;
            }

            //this.LoadComplete += new EventHandler(BaseWebPage_EventComplete);
            //this.DataBinding += new EventHandler(BaseWebPage_EventComplete);
            //this.Disposed += new EventHandler(BaseWebPage_EventComplete);
            //this.Init += new EventHandler(BaseWebPage_EventComplete);
            //this.InitComplete +=new EventHandler(BaseWebPage_EventComplete);
            //this.Load += new EventHandler(BaseWebPage_EventComplete);
            //this.PreInit += new EventHandler(BaseWebPage_EventComplete);
            //this.PreLoad += new EventHandler(BaseWebPage_EventComplete);
        }
        #endregion

        #region PRIVATE
        private readonly AspSystemManager _aspSystemManager;
        private readonly BaseIocContainer _baseContainer;
        private readonly IList<Control> _manualRenderList = new List<Control>();
        private void RenderControlToCache(Control pControl)
        {
            //render and save for later (jquery will .load it when needed)
            string html = RenderControl(pControl.Controls[1]);
            var cacheId = Guid.NewGuid().ToString();
            Cache.Add(cacheId, html, null, System.Web.Caching.Cache.NoAbsoluteExpiration, new TimeSpan(0, 0, 60), System.Web.Caching.CacheItemPriority.Default, null);

            pControl.Controls[1].Visible = false; //prevent from being rendered with the page
            ((HtmlGenericControl)pControl).Attributes["CachId"] = cacheId; //mark the control with cacheid(guid)
        }
        #endregion

        #region PAGE LIFE-CYCLE
        protected override void OnInit(EventArgs e)
        {
            var cacheGuid = Request.QueryString["callback"] ?? string.Empty;
            if (cacheGuid != string.Empty)
            {
                var html = Cache.Get(cacheGuid).ToString();
                Response.Write(html);
                Response.End();
            }
            else
            {
                base.OnInit(e);
            }
        }
        protected override void OnLoad(EventArgs e)
        {
            //if (!IsPostBack)
            //{
                //this application may have a js requirement?
                var applicationScript = _aspSystemManager.GetApplicationScript();
                if (!string.IsNullOrEmpty(applicationScript))
                {
                    RegisterPageScript(applicationScript);
                }

                //as might this page
                if (!string.IsNullOrEmpty(PageProfile.PageScript))
                {
                    RegisterPageScript(PageProfile.PageScript);
                }
                //and maybe a wcf 
                if (!string.IsNullOrEmpty(PageProfile.PageService))
                {
                    RegisterPageService(PageProfile.PageService);
                }  
            //}
            base.OnLoad(e);
            DataBind();
        }
        protected override void OnPreRender(EventArgs e)
        {
            if (!this.IsPostBack)
            {
                foreach(var ctl in _manualRenderList)
                {
                    this.RenderControlToCache(ctl);
                }
            }
            base.OnPreRender(e);
        }
        #endregion

        /// <summary>
        /// listview UiItem created... fixup as necessary
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        /// <remarks>
        /// A page can be constructed by rendering UIDataItem(s) that come from
        /// configuration.  Each of those items cause this event to fire.
        /// </remarks>
        protected void lvBizworksItemCreated(object sender, ListViewItemEventArgs e)
        {
            var uiControl = e.Item.FindControl("ucRenderUiItem");
            var uiItem = (UiDataItem)e.Item.DataItem;
            switch (uiItem.DataControl)
            {
                case "DropDownList":
                    var ddl = (DropDownList)uiControl.FindControl("BizworksDropDownList");
                    //virtual method.. hoping presenter overrides it and provides values
                    ddl.DataSource = PagePresenter.GetDropDownList(uiItem.DataName);
                    break;
                case "TextBox":
                    var tb = (TextBox)uiControl.FindControl("BizworksTextBox");
                    break;
                // TODO: what does this mean?
                default: break;
            }
        }

        #region PROPERTIES
        public PageProfile PageProfile { get; private set; }
        /// <summary>
        /// Reference to the Bizworks[Presenter] for this UI.Page
        /// </summary>
        /// <remarks>
        /// The PagePresenter provides all application interaction for the page including data access and other various application related methods.  If 
        /// a PresenterDataSource(s) is/are being used on the page, they will be wired to this reference at runtime.   This must remain Public so 
        /// that the PresenterDataSource can find it with reflection.
        /// NOTE: PresenterDataSource is sensitive (reflection) to the actual name of this property
        /// </remarks>
        public TPagePresenter PagePresenter { get; private set; }
        #endregion

        /// <summary>
        /// register additional presenter
        /// </summary>
        /// <param name="pPresenter"></param>
        protected void RegisterOtherPresenter(BasePresenter pPresenter)
        {
            pPresenter.RaiseNavigationEvent += PagePresenterRaiseNavigationEvent;
        }
        /// <summary>
        /// handle request from presenter to goto another page
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void PagePresenterRaiseNavigationEvent(object sender, NavigatonEventArgs e)
        {
            Response.Redirect(e.PageAspx, false);
        }
        /// <summary>
        /// register page service (.svc)
        /// </summary>
        /// <param name="pServiceFile"></param>
        protected virtual void RegisterPageService(string pServiceFile)
        {
            var scriptManager = ScriptManager.GetCurrent(this);
            if (scriptManager != null)
            {
                scriptManager.Services.Add(new ServiceReference(pServiceFile));
            }
        }
        /// <summary>
        /// register page script (.js)
        /// </summary>
        /// <param name="pServiceFile"></param>
        protected virtual void RegisterPageScript(string pServiceFile)
        {
            var scriptManager = ScriptManager.GetCurrent(this);
            if (scriptManager != null)
            {
                scriptManager.Scripts.Add(new ScriptReference(pServiceFile));
            }
        }
        /// <summary>
        /// find control on this page
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="controls"></param>
        /// <returns></returns>
        protected static T FindControl<T>(ControlCollection controls) where T : class
        {
            T found = default(T);
            if (controls != null && controls.Count > 0)
            {
                for (int i = 0; i < controls.Count; i++)
                {
                    if (controls[i] is T)
                    {
                        found = controls[i] as T;
                        break;
                    }
                    found = FindControl<T>(controls[i].Controls);
                }
            }
            return found;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="control"></param>
        /// <returns></returns>
        /// <remarks>
        /// If you get exception "RegisterForEventValidation can only be called during Render()" try
        /// putting EnableEventValidation = "false" in the page directive.
        /// </remarks>
        protected static string RenderControl(Control control)
        {
            var tw = new StringWriter();
            var writer = new Html32TextWriter(tw);
            // NOTE: if this causes an error (RegisterForEventValidation), 
            control.RenderControl(writer);
            writer.Close();
            return tw.ToString();
        }
        /// <summary>
        /// Verify Rendering In Server Form
        /// </summary>
        /// <param name="control"></param>
        /// <remarks>
        /// This is necessary to keep the .RenderControl method (above) from throwing an exception.
        /// </remarks>
        public override void VerifyRenderingInServerForm(Control control)
        {
            return;
        }
        /// <summary>
        /// Set manual render 
        /// </summary>
        /// <param name="pControl">control to manually render</param>
        /// <remarks>
        /// this method allows derived classes to tell which controls should be rendered 
        /// outside of the normal page life-cycle. these controls will be rendered
        /// to a cache and later retrieved by the client-side script when/if needed.
        /// </remarks>
        protected void SetManualRender(Control pControl)
        {
            this._manualRenderList.Add(pControl);
        }
    }
}
